"""
---------------------------------------------------------------------------------------------------
 How to Access the LP DAAC Data Pool with Python
 The following Python code example demonstrates how to configure a connection to download data from
 an Earthdata Login enabled server, specifically the LP DAAC's Data Pool.
---------------------------------------------------------------------------------------------------
 Author: Cole Krehbiel
 Last Updated: 11/20/2018
---------------------------------------------------------------------------------------------------
"""
import argparse
import os
import time
from getpass import getpass
from netrc import netrc

# Load necessary packages into Python
from subprocess import Popen

import requests

# ----------------------------------USER-DEFINED VARIABLES--------------------------------------- #
# Set up command line arguments
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("-dir", "--directory", required=True, help="Specify directory to save files to")
parser.add_argument(
    "-f",
    "--files",
    required=True,
    help="A single granule URL, or the location of textfile containing granule URLs",
)
args = parser.parse_args()

saveDir = args.directory  # Set local directory to download to
# Define file(s) to download from the LP DAAC Data Pool
files = args.files
prompts = [
    "Enter NASA Earthdata Login Username \n(or create an account at urs.earthdata.nasa.gov): ",
    "Enter NASA Earthdata Login Password: ",
]

# ---------------------------------SET UP WORKSPACE---------------------------------------------- #
# Create a list of files to download based on input type of files above
if files.endswith(".txt"):
    fileList = open(files).readlines()  # If input is textfile w file URLs
elif isinstance(files, str):
    fileList = [files]  # If input is a single file

# Generalize download directory
if saveDir[-1] != "/" and saveDir[-1] != "\\":
    saveDir = saveDir.strip("'").strip('"') + os.sep
urs = "urs.earthdata.nasa.gov"  # Address to call for authentication

# --------------------------------AUTHENTICATION CONFIGURATION----------------------------------- #
# Determine if netrc file exists, and if so, if it includes NASA Earthdata Login Credentials
try:
    netrcDir = os.path.expanduser("~/.netrc")
    netrc(netrcDir).authenticators(urs)[0]

# Below, create a netrc file and prompt user for NASA Earthdata Login Username and Password
except FileNotFoundError:
    homeDir = os.path.expanduser("~")
    Popen(
        "touch {0}.netrc | chmod og-rw {0}.netrc | echo machine {1} >> {0}.netrc".format(
            homeDir + os.sep, urs
        ),
        shell=True,
    )
    Popen(f"echo login {getpass(prompt=prompts[0])} >> {homeDir + os.sep}.netrc", shell=True)
    Popen(
        f"echo password {getpass(prompt=prompts[1])} >> {homeDir + os.sep}.netrc",
        shell=True,
    )

# Determine OS and edit netrc file if it exists but is not set up for NASA Earthdata Login
except TypeError:
    homeDir = os.path.expanduser("~")
    Popen(f"echo machine {urs} >> {homeDir + os.sep}.netrc", shell=True)
    Popen(f"echo login {getpass(prompt=prompts[0])} >> {homeDir + os.sep}.netrc", shell=True)
    Popen(
        f"echo password {getpass(prompt=prompts[1])} >> {homeDir + os.sep}.netrc",
        shell=True,
    )

# Delay for up to 1 minute to allow user to submit username and password before continuing
tries = 0
while tries < 30:
    try:
        netrc(netrcDir).authenticators(urs)[2]
    except Exception:
        time.sleep(2.0)
    tries += 1

# -----------------------------------------DOWNLOAD FILE(S)-------------------------------------- #
# Loop through and download all files to the directory specified above, and keeping same filenames
for f in fileList:
    if not os.path.exists(saveDir):
        os.makedirs(saveDir)
    saveName = os.path.join(saveDir, f.split("/")[-1].strip())

    # Create and submit request and download file
    with requests.get(
        f.strip(),
        stream=True,
        auth=(netrc(netrcDir).authenticators(urs)[0], netrc(netrcDir).authenticators(urs)[2]),
    ) as response:
        if response.status_code != 200:
            print(
                "{} not downloaded. Verify that your username and password are correct in {}".format(
                    f.split("/")[-1].strip(), netrcDir
                )
            )
        else:
            response.raw.decode_content = True
            content = response.raw
            with open(saveName, "wb") as d:
                while True:
                    chunk = content.read(16 * 1024)
                    if not chunk:
                        break
                    d.write(chunk)
            print(f"Downloaded file: {saveName}")


def download_file(saveDir, netrcDir):
    if not os.path.exists(saveDir):
        os.makedirs(saveDir)
    # saveName = os.path.join(saveDir, f.split("/")[-1].strip())
